//  
//  Copyright (C) 2009 Bartek thindil Jasicki
// 
//  This file is part of Grubng
// 
//  Grubng is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 
using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Net;
using System.Runtime.InteropServices;


namespace urlsdb
{
	/// <summary>
	/// Main program class
	/// </summary>
	class MainClass
	{
		/// <summary>
		/// Main program function. Add, delete URL's from database and Solr, create workunits, config file
		/// and URL's database.
		/// </summary>
		/// <param name="args">
		/// A <see cref="System.String"/> option to program. Run program without option for get help.
		/// </param>
		public static void Main(string[] args)
		{
			SetProcessName("grub-urlsdb");
			//If no arguments for program, show help
			if (args.Length == 0)
			{
				Console.WriteLine("createconf - create configuration file for program");
				Console.WriteLine("updateconf - update configuration file for program");
				Console.WriteLine("createdb - create database");
				Console.WriteLine("filldb [FILENAME] - put URL's to database from FILENAME file");
				Console.WriteLine("backup [FILENAME] - create database backup to FILENAME file (only URL's)");
				Console.WriteLine("addurl [URL] - add URL to database");
				Console.WriteLine("delurl [URL] - delete URL from database and index");
				Console.WriteLine("delurls [FILENAME] - delete URL's (list from FILENAME) from database and index");
				Console.WriteLine("createwu - create workunits from URL's from database.");
				Console.WriteLine("maintain - delete URL's from database (with HTTP status codes: 301 and 400-406), optimize tables and optimize Solr");
				Console.WriteLine("optimizedb - optimize and flush database");
				Console.WriteLine("optimizesolr - optimize Solr index (if Solr is enabled)");
				return;
			}
			//Create configuration file
			if (args[0] == "createconf")
			{
				Console.Write("Creating configuration file ... ");
				Config.CreateConfig();				
				Console.WriteLine("done.");
				Console.WriteLine("Please now edit it for set correct settings");
				return;
			}
			//Update configuration file
			if (args[0] == "updateconf")
			{
				Console.Write("Updating configuration file ... ");
				Config.UpdateConfig();
				Console.WriteLine("done.");
				Console.WriteLine("Please now edit it for set correct settings");
				return;
			}
			//Create database
			if (args[0] == "createdb")
			{
				Console.Write("Creating database ... ");
				Database.CreateDB();
				Console.WriteLine("done.");
				return;
			}
			Database db = new Database();
			db.Open();
			//Backup URL's to plain text file
			if (args[0] == "backup")
			{
				if (args.Length == 1)
				{
					Console.WriteLine("Please enter backup file name.");
					return;
				}
				Console.Write("Creating database backup ...");
				int offset = 0;
				//string urls;
				StreamWriter writer = new StreamWriter(args[1]);
				StringBuilder urls2 = new StringBuilder();
				while (true)
				{
					urls2.Append(db.SelectURLs(offset));
					if (urls2.Length == 0)
					{
						break;
					}
					urls2.Append(Environment.NewLine);
					writer.Write(urls2);
					writer.Flush();
					urls2.Remove(0, urls2.Length);
					offset += 1000000;
				}
				writer.Close();
				writer.Dispose();
				Console.WriteLine("done.");
			}
			ParseURLs parseurl = new ParseURLs();
			int amount = 0;
			string record;
			//Insert URL's from file to database
			if (args[0] == "filldb")
			{
				if (args.Length == 1)
				{
					Console.WriteLine("Please enter file name with urls.");
					return;
				}
				if (!File.Exists(args[1]))
				{
					Console.WriteLine("File {0} not exist.", args[1]);
					return;
				}
				StreamReader reader = new StreamReader(args[1]);
				int i = 0;
				while (!reader.EndOfStream)
				{
					record = parseurl.ParseURL(reader.ReadLine());
					if (record.Length > 0)
					{
						amount += db.InstertURL(parseurl.GetHash(record), record);
					}
					i ++;
				}
				reader.Close();
				reader.Dispose();
				Console.WriteLine("{0} URL's was inserted from {1} URL's.", amount.ToString(), i.ToString());
			}
			//Add single URL to database
			if (args[0] == "addurl")
			{
				if (args.Length == 1)
				{
					Console.WriteLine("Please enter URL to add.");
					return;
				}
				record = parseurl.ParseURL(args[1]);
				if (record.Length > 0)
				{
					amount = db.InstertURL(parseurl.GetHash(record), record);
				}
				if (amount == 0)
				{
					Console.WriteLine("URL already exist in database.");
				}
				else
				{
					Console.WriteLine("URL added to database.");
				}
			}
			bool solrenabled = false;
			if (Config.ReadConfig("/configuration/enablesolr") == "Y")
			{
				solrenabled = true;
			}
			//Delete single URL from database
			if (args[0] == "delurl")
			{
				if (args.Length == 1)
				{
					Console.WriteLine("Please enter URL to delete.");
					return;
				}
				amount = db.DeleteURL(parseurl.GetHash(args[1]));
				if (solrenabled)
				{
					string solrcommand = "<delete><id>http://" + parseurl.ParseURL(args[1]) + "</id></delete>";
					solrcommand = Regex.Replace(solrcommand, @"[\p{IsC}]", String.Empty);
					MainClass.SendToSolr(solrcommand);
					solrcommand = "<commit/>";
					MainClass.SendToSolr(solrcommand);
				}
				if (amount == 0)
				{
					Console.WriteLine("URL not exist in database.");
				}
				else
				{
					Console.WriteLine("URL deleted from database.");
				}
			}
			//Delete URL's from file from database
			if (args[0] == "delurls")
			{
				if (args.Length == 1)
				{
					Console.WriteLine("Please enter file name with urls.");
					return;
				}
				if (!File.Exists(args[1]))
				{
					Console.WriteLine("File {0} not exist.", args[1]);
					return;
				}
				StreamReader reader = new StreamReader(args[1]);
				string solrcommand = "<delete>";
				int i = 0;
				while (!reader.EndOfStream)
				{
					record = reader.ReadLine();
					amount += db.DeleteURL(parseurl.GetHash(record));
					solrcommand += "<id>http://" + parseurl.ParseURL(record) + "</id>";
					i ++;
				}
				reader.Close();
				reader.Dispose();
				if (solrenabled)
				{
					solrcommand += "</delete>";
					solrcommand = Regex.Replace(solrcommand, @"[\p{IsC}]", String.Empty);
					MainClass.SendToSolr(solrcommand);
					solrcommand = "<commit/>";
					MainClass.SendToSolr(solrcommand);
				}
				Console.WriteLine("{0} URL's was deleted from {1} URL's.", amount.ToString(), i.ToString());
			}
			//Create workunits from URL's from database
			if (args[0] == "createwu")
			{
				string urls, path, host;
				string wupass = Config.ReadConfig("/configuration/workunitspassword");
				string useragent = Config.ReadConfig("/configuration/useragent");
				int urlsamount = Convert.ToInt32(Config.ReadConfig("/configuration/urlsamount"));
				string httpversion = Config.ReadConfig("/configuration/httpversion");
				string accept = Config.ReadConfig("/configuration/accept");
				string[] urlsArray, urlparts;
				int offset = 0;
				urls = db.SelectURLs(-1);
				if (urls.Length == 0)
				{
					return;
				}
				string workunitsdirectory = Config.ReadConfig("/configuration/workunitsdirectory");
				string tempwu = workunitsdirectory + "wu.temp";
				StreamWriter writer = new StreamWriter(tempwu);
				DateTime modtime;
				urlsArray = urls.Split(new char[] {'\n'});
				StringBuilder key = new StringBuilder();
				System.Security.Cryptography.SHA1Managed SHhash = new System.Security.Cryptography.SHA1Managed();
				byte[] PureHash = Encoding.UTF8.GetBytes(wupass);
				byte[] HashValue = SHhash.ComputeHash(PureHash);
				foreach(byte b in HashValue) 
				{
					key.Append(String.Format("{0:x2}", b));
				}
				foreach (string url in urlsArray)
				{
					urlparts = url.Split(new char[] {' '});
					if (urlparts.Length != 2)
					{
						continue;
					}
					if (urlparts[0].IndexOf('/') == -1)
					{
						path = "/";
						host = urlparts[0];
					}
					else
					{
						path = urlparts[0].Substring(url.IndexOf('/'));
						host = urlparts[0].Remove(url.IndexOf('/'));
					}
					writer.Write("GET " + path + " HTTP/" + httpversion +"\r\n");
					writer.Write("Host: " + host + "\r\n");
					writer.Write("User-Agent: " + useragent + "\r\n");
					writer.Write("Accept: " + accept +"\r\n");
					if (urlparts[1] != "0")
					{
						modtime = new DateTime(Convert.ToInt32(urlparts[1].Substring(0, 4)), 
						                       Convert.ToInt32(urlparts[1].Substring(4, 2)),
						                       Convert.ToInt32(urlparts[1].Substring(6, 2)),
						                       Convert.ToInt32(urlparts[1].Substring(8, 2)),
						                       Convert.ToInt32(urlparts[1].Substring(10, 2)),
						                       Convert.ToInt32(urlparts[1].Substring(12)));
						writer.Write("If-Modified-Since: " + modtime.ToString("r") + "\r\n");
					}
					writer.Write("\r\n");
					writer.Flush();
					PureHash = Encoding.UTF8.GetBytes(key.ToString() + " " + host + " " + path);
					key.Remove(0, key.Length);
					HashValue = SHhash.ComputeHash(PureHash);
					foreach(byte b in HashValue) 
					{
						key.Append(String.Format("{0:x2}", b));
					}
					offset ++;
					if (offset == urlsamount)
					{
						writer.Close();
						writer.Dispose();
						File.Copy(tempwu, workunitsdirectory + key.ToString() + ".wu1", true);
						File.Delete(tempwu);
						key.Remove(0, key.Length);
						PureHash = Encoding.UTF8.GetBytes(wupass);
						HashValue = SHhash.ComputeHash(PureHash);
						foreach(byte b in HashValue) 
						{
							key.Append(String.Format("{0:x2}", b));
						}
						writer = new StreamWriter(tempwu);
						offset = 0;
					}
				}
				writer.Close();
				writer.Dispose();
				File.Delete(tempwu);
			}
			db.Close();
			//Maintenance work on URL's database and Solr
			if (args[0] == "maintain")
			{
				string[] codes = new string[] {"204", "301", "400", "401", "402", "403", "404", "405", "406", "408", "503"};
				foreach (string code in codes)
				{
					Console.WriteLine("Deleting URL's with HTTP status code {0}", code);
					db.Open();
					amount = db.CleanURLs(code, solrenabled);
					db.Close();
					Console.WriteLine("{0} URL's was deleted.", amount);
				}
			}
			//Optimize and flush database
			if ((args[0] == "optimizedb") || (args[0] == "maintain"))
			{
				Console.Write("Optimizing database ... ");
				db.Open();
				db.OptimizeDB();
				db.Close();
				Console.WriteLine("done.");
			}
			//Optimize Solr index
			if ((args[0] == "optimizesolr") || (args[0] == "maintain"))
			{
				if (!solrenabled)
				{
					return;
				}
				Console.Write("Optimizing Solr ... ");
				MainClass.SendToSolr("<optimize/>");
				Console.WriteLine("done.");
			}
		}
		
		/// <summary>
		/// Function set process name on Unix systems. Code borrowed from 
		/// http://abock.org/2006/02/09/changing-process-name-in-mono/
		/// </summary>
		/// <param name="name">
		/// A <see cref="System.String"/> new name of process
		/// </param>
		public static void SetProcessName (string name)
		{
			try 
			{
				if (NativeMethods.prctl (15 /* PR_SET_NAME */, Encoding.ASCII.GetBytes (name + "\0"),
				           IntPtr.Zero, IntPtr.Zero, IntPtr.Zero) != 0) {
					throw new ApplicationException ("Error setting process name: " +
					                                Mono.Unix.Native.Stdlib.GetLastError ());
				}
			} 
			catch (EntryPointNotFoundException) 
			{
				NativeMethods.setproctitle (Encoding.ASCII.GetBytes ("%s\0"),
				              Encoding.ASCII.GetBytes (name + "\0"));
			}
		}
		
		/// <summary>
		/// Function send XML Update command to Solr
		/// </summary>
		/// <param name="command">
		/// A <see cref="System.String"/> XML Update command to send
		/// </param>
		public static void SendToSolr(string command)
		{
			HttpWebRequest request = (HttpWebRequest)WebRequest.Create(Config.ReadConfig("/configuration/solraddress"));
			request.Proxy = null;
			request.Credentials = new NetworkCredential(Config.ReadConfig("/configuration/solrusername"), 
			                                            Config.ReadConfig("configuration/solrpassword"));
			request.PreAuthenticate = true;
			request.ContentType = "text/xml";
			request.Method = "POST";
			request.Timeout = 1000000000;
			Stream streamw1 = request.GetRequestStream();
			byte[] buffer = Encoding.UTF8.GetBytes(command);
			streamw1.Write(buffer, 0, buffer.Length);
			streamw1.Close();
			streamw1.Dispose();
			try
			{
				HttpWebResponse response = (HttpWebResponse)request.GetResponse();
				if (response != null)
				{
					response.Close();
				}
			}
			catch (WebException e)
			{
				if (e.Response != null)
				{
					Stream stream = e.Response.GetResponseStream();
					FileStream errorfile = new FileStream("error.log", FileMode.Append);
					string tdate = DateTime.Now.ToString("yyyy/MM/dd/HH:mm:ss");
					buffer = Encoding.UTF8.GetBytes(tdate + " Solr error:" + Environment.NewLine);
					errorfile.Write(buffer, 0, buffer.Length);
					errorfile.Flush();
					buffer = new byte[1024];
					int bytesRead = 0;
					while ((bytesRead = stream.Read(buffer, 0, buffer.Length)) != 0)
					{
						errorfile.Write(buffer, 0, bytesRead);
						errorfile.Flush();
					}
					stream.Close();
					stream.Dispose();
					errorfile.Close();
					errorfile.Dispose();
					e.Response.Close();
				}
			}
		}
	}
	
	/// <summary>
	/// Provide p/invoke for native Unix methods
	/// </summary>
	internal static class NativeMethods
	{
		[DllImport ("libc")] // GNU/Linux
		public static extern int prctl (int option, byte [] arg2, IntPtr arg3,
		                                 IntPtr arg4, IntPtr arg5);

		[DllImport ("libc")] // BSD
		public static extern void setproctitle (byte [] fmt, byte [] str_arg);
	}
}